/*
 *
 * LineByLinePrinter (line-by-line-printer.js)
 * Author: rtfpessoa
 *
 */

(function() {
  var minimatch = require('minimatch');
  var diffParser = require('./diff-parser.js').DiffParser;
  var printerUtils = require('./printer-utils.js').PrinterUtils;
  var utils = require('./utils.js').Utils;
  var Rematch = require('./rematch.js').Rematch;

  var hoganUtils;

  var genericTemplatesPath = 'generic';
  var baseTemplatesPath = 'line-by-line';
  var iconsBaseTemplatesPath = 'icon';
  var tagsBaseTemplatesPath = 'tag';

  function LineByLinePrinter(config) {
    this.config = config;

    var HoganJsUtils = require('./hoganjs-utils.js').HoganJsUtils;
    hoganUtils = new HoganJsUtils(config);
  }

  LineByLinePrinter.prototype.makeFileDiffHtml = function(file, diffs) {
    if (this.config.renderNothingWhenEmpty && file.blocks && !file.blocks.length) return '';

    var fileDiffTemplate = hoganUtils.template(baseTemplatesPath, 'file-diff');
    var filePathTemplate = hoganUtils.template(genericTemplatesPath, 'file-path');
    var fileIconTemplate = hoganUtils.template(iconsBaseTemplatesPath, 'file');
    var fileTagTemplate = hoganUtils.template(tagsBaseTemplatesPath, printerUtils.getFileTypeIcon(file));

    return fileDiffTemplate.render({
      file: file,
      fileHtmlId: printerUtils.getHtmlId(file),
      diffs: diffs,
      filePath: filePathTemplate.render({
        fileDiffName: printerUtils.getDiffName(file)
      }, {
        fileIcon: fileIconTemplate,
        fileTag: fileTagTemplate
      })
    });
  };

  LineByLinePrinter.prototype.makeLineByLineHtmlWrapper = function(content) {
    return hoganUtils.render(genericTemplatesPath, 'wrapper', {'content': content});
  };

  LineByLinePrinter.prototype.generateLineByLineJsonHtml = function(diffFiles, coverage, coverageFilePatterns, workspace) {
    var that = this;
    that.coverageFilePatterns = coverageFilePatterns;
    that.workspace = workspace;
    var htmlDiffs = diffFiles.map(function(file) {
      var diffs;
      if (file.blocks.length) {
        diffs = that._generateFileHtml(file, coverage);
      } else {
        diffs = that._generateEmptyDiff();
      }
      return that.makeFileDiffHtml(file, diffs);
    });

    return this.makeLineByLineHtmlWrapper(htmlDiffs.join('\n'));
  };

  var matcher = Rematch.rematch(function(a, b) {
    var amod = a.content.substr(1);
    var bmod = b.content.substr(1);

    return Rematch.distance(amod, bmod);
  });

  LineByLinePrinter.prototype.makeColumnLineNumberHtml = function(block) {
    return hoganUtils.render(genericTemplatesPath, 'column-line-number', {
      diffParser: diffParser,
      blockHeader: utils.escape(block.header),
      lineClass: 'd2h-code-linenumber',
      contentClass: 'd2h-code-line'
    });
  };

  LineByLinePrinter.prototype._generateFileHtml = function(file, coverage) {
    var that = this;

    function getNewLineCoverageStatus(file, newLine, coverage) {
      // line coverage status: 0: uncovered 1: covered 2: non-executable
      if (that._isExcludeFile(file) || !that._isExecutableFile(file.newName)) {
        newLine.coverageStatus = 2;
        global.diffStatementNewLines += 1;
        global.nonExecutableDiffStatementNewLines += 1;
        return;
      }
      if (coverage) {
        newLine.coverageStatus = 2;
        const fileName = that.workspace ? that.workspace + file.newName : file.newName;
        const fileCoverage = coverage[fileName];
        if (!fileCoverage) {
          newLine.coverageStatus = 0; // mark line uncovered if coverage data not found for current file
          that._increaseDiffStatementNewLines(file, newLine.content, newLine.coverageStatus);
          return;
        }
        const newLineNewNumber = newLine.newNumber;
        const statementMap = fileCoverage.statementMap;
        const s = fileCoverage.s;
        let lineIsStatement = false;
        for (const blockNo in statementMap) {
          const block = statementMap[blockNo];
          if (block.start.line <= newLineNewNumber && block.end.line >= newLineNewNumber) { // line is statement
            lineIsStatement = true;
            newLine.coverageStatus = s[blockNo] > 0 ? 1 : 0;
            break;
          }
        }
        if (!lineIsStatement) {
          newLine.coverageStatus = 2;
        }
        that._increaseDiffStatementNewLines(file, newLine.content, newLine.coverageStatus);
      }
    }

    return file.blocks.map(function(block) {
      var lines = that.makeColumnLineNumberHtml(block);
      var oldLines = [];
      var newLines = [];

      function processChangeBlock() {
        var matches;
        var insertType;
        var deleteType;

        var comparisons = oldLines.length * newLines.length;

        var maxLineSizeInBlock = Math.max.apply(null,
          [0].concat((oldLines.concat(newLines)).map(
            function(elem) {
              return elem.content.length;
            }
          )));

        var doMatching = comparisons < that.config.matchingMaxComparisons &&
          maxLineSizeInBlock < that.config.maxLineSizeInBlockForComparison &&
          (that.config.matching === 'lines' || that.config.matching === 'words');

        if (doMatching) {
          matches = matcher(oldLines, newLines);
          insertType = diffParser.LINE_TYPE.INSERT_CHANGES;
          deleteType = diffParser.LINE_TYPE.DELETE_CHANGES;
        } else {
          matches = [[oldLines, newLines]];
          insertType = diffParser.LINE_TYPE.INSERTS;
          deleteType = diffParser.LINE_TYPE.DELETES;
        }

        matches.forEach(function(match) {
          oldLines = match[0];
          newLines = match[1];

          var processedOldLines = [];
          var processedNewLines = [];

          var common = Math.min(oldLines.length, newLines.length);

          var oldLine, newLine;
          for (var j = 0; j < common; j++) {
            oldLine = oldLines[j];
            newLine = newLines[j];

            that.config.isCombined = file.isCombined;
            var diff = printerUtils.diffHighlight(oldLine.content, newLine.content, that.config);

            processedOldLines +=
              that.makeLineHtml(file.isCombined, deleteType, oldLine.oldNumber, oldLine.newNumber,
                diff.first.line, diff.first.prefix, oldLine.coverageStatus);

            processedNewLines +=
              that.makeLineHtml(file.isCombined, insertType, newLine.oldNumber, newLine.newNumber,
                diff.second.line, diff.second.prefix, newLine.coverageStatus);
          }

          lines += processedOldLines + processedNewLines;
          lines += that._processLines(file.isCombined, oldLines.slice(common), newLines.slice(common));
        });

        oldLines = [];
        newLines = [];
      }

      for (var i = 0; i < block.lines.length; i++) {
        var line = block.lines[i];
        var escapedLine = utils.escape(line.content);

        if (line.type !== diffParser.LINE_TYPE.INSERTS &&
          (newLines.length > 0 || (line.type !== diffParser.LINE_TYPE.DELETES && oldLines.length > 0))) {
          processChangeBlock();
        }

        if (line.type === diffParser.LINE_TYPE.CONTEXT) {
          lines += that.makeLineHtml(file.isCombined, line.type, line.oldNumber, line.newNumber, escapedLine, '', line.coverageStatus);
        } else if (line.type === diffParser.LINE_TYPE.INSERTS && !oldLines.length) {
          getNewLineCoverageStatus(file, line, coverage);
          lines += that.makeLineHtml(file.isCombined, line.type, line.oldNumber, line.newNumber, escapedLine, '', line.coverageStatus);
        } else if (line.type === diffParser.LINE_TYPE.DELETES) {
          oldLines.push(line);
        } else if (line.type === diffParser.LINE_TYPE.INSERTS && Boolean(oldLines.length)) {
          getNewLineCoverageStatus(file, line, coverage);
          newLines.push(line);
        } else {
          console.error('Unknown state in html line-by-line generator');
          processChangeBlock();
        }
      }

      processChangeBlock();

      return lines;
    }).join('\n');
  };

  LineByLinePrinter.prototype._processLines = function(isCombined, oldLines, newLines) {
    var lines = '';

    for (var i = 0; i < oldLines.length; i++) {
      var oldLine = oldLines[i];
      var oldEscapedLine = utils.escape(oldLine.content);
      lines += this.makeLineHtml(isCombined, oldLine.type, oldLine.oldNumber, oldLine.newNumber, oldEscapedLine);
    }

    for (var j = 0; j < newLines.length; j++) {
      var newLine = newLines[j];
      var newEscapedLine = utils.escape(newLine.content);
      lines += this.makeLineHtml(isCombined, newLine.type, newLine.oldNumber, newLine.newNumber, newEscapedLine, '', newLine.coverageStatus);
    }

    return lines;
  };

  LineByLinePrinter.prototype.makeLineHtml = function(isCombined, type, oldNumber, newNumber, content, possiblePrefix, newLineCoverageStatus) {
    let newNumberContent = utils.valueOrEmpty(newNumber);
    if (newLineCoverageStatus !== undefined) {
      if (type === 'd2h-ins') {
        switch (newLineCoverageStatus) {
          case 0:
            newNumberContent = `${utils.valueOrEmpty(newNumber)} 🔴`; // uncovered
            break;
          case 1:
            newNumberContent = `${utils.valueOrEmpty(newNumber)} 🔵`; // covered
            break;
          case 2:
            newNumberContent = utils.valueOrEmpty(newNumber); // non-executable line
            break;
          default:
            newNumberContent = utils.valueOrEmpty(newNumber);
        }
      }
    }
    var lineNumberTemplate = hoganUtils.render(baseTemplatesPath, 'numbers', {
      oldNumber: utils.valueOrEmpty(oldNumber),
      newNumber: newNumberContent
    });

    var lineWithoutPrefix = content;
    var prefix = possiblePrefix;

    if (!prefix) {
      var lineWithPrefix = printerUtils.separatePrefix(isCombined, content);
      prefix = lineWithPrefix.prefix;
      lineWithoutPrefix = lineWithPrefix.line;
    }

    if (prefix === ' ') {
      prefix = '&nbsp;';
    }

    return hoganUtils.render(genericTemplatesPath, 'line',
      {
        type: type,
        lineClass: 'd2h-code-linenumber',
        contentClass: 'd2h-code-line',
        prefix: prefix,
        content: lineWithoutPrefix,
        lineNumber: lineNumberTemplate
      });
  };

  LineByLinePrinter.prototype._generateEmptyDiff = function() {
    return hoganUtils.render(genericTemplatesPath, 'empty-diff', {
      contentClass: 'd2h-code-line',
      diffParser: diffParser
    });
  };

  LineByLinePrinter.prototype._increaseDiffStatementNewLines = function(file, line, coverageStatus) {
    global.diffStatementNewLines += 1;
    if (coverageStatus === 0) {
      global.uncoveredDiffStatementNewLines += 1;
    } else if (coverageStatus === 1) {
      global.coveredDiffStatementNewLines += 1;
    } else if (coverageStatus === 2) {
      global.nonExecutableDiffStatementNewLines += 1;
    }
  };

  LineByLinePrinter.prototype._isExcludeFile = function(file) {
    if (!this.coverageFilePatterns) {
      return false;
    }
    var exclude = true;
    for (var i = 0; i < this.coverageFilePatterns.length; i += 1) {
      var group = this.coverageFilePatterns[i];
      for (var j = 0; j < group.length; j += 1) {
        if (!minimatch(file.newName, group[j])) {
          break;
        }
        if (j === group.length - 1) {
          exclude = false;
        }
      }
      if (!exclude) {
        break;
      }
    }
    return exclude;
  };

  LineByLinePrinter.prototype._isExecutableFile = function(fileName) {
    if (fileName.indexOf('test') !== -1) {
      return false;
    }
    const extensions = [
      'js',
      'ts',
      'jsx',
      'tsx',
      'cpp',
      'c',
      'java',
      'kt',
      'swift'
    ];
    for (let i = 0; i < extensions.length; i += 1) {
      const ext = extensions[i];
      if (fileName.endsWith(ext)) {
        return true;
      }
    }
    return false;
  };

  module.exports.LineByLinePrinter = LineByLinePrinter;
})();
